#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2021, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
'''

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Union,
)

# External imports
from typing_extensions import Literal, TypedDict

## Bokeh imports
if TYPE_CHECKING:
    from ..core.has_props import ModelDef
    from ..core.types import (
        ID,
        Ref,
        ReferenceJson,
        Unknown,
    )
    from ..models.sources import DataDict

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = ()

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

Patch = Any # TODO

Patches = Dict[str, List[Patch]]

class ModelChanged(TypedDict):
    kind: Literal["ModelChanged"]
    model: Ref
    attr: str
    new: Unknown
    hint: DocumentPatched | None  # type: ignore[misc] # https://github.com/python/mypy/issues/731

class MessageSent(TypedDict):
    kind: Literal["MessageSent"]
    msg_type: str
    msg_data: Unknown | None

class TitleChanged(TypedDict):
    kind: Literal["TitleChanged"]
    title: str

class RootAdded(TypedDict):
    kind: Literal["RootAdded"]
    model: Ref

class RootRemoved(TypedDict):
    kind: Literal["RootRemoved"]
    model: Ref

class ColumnDataChanged(TypedDict):
    kind: Literal["ColumnDataChanged"]
    column_source: Ref
    cols: List[str] | None
    new: Unknown

class ColumnsStreamed(TypedDict):
    kind: Literal["ColumnsStreamed"]
    column_source: Ref
    data: DataDict
    rollover: int | None

class ColumnsPatched(TypedDict):
    kind: Literal["ColumnsPatched"]
    column_source: Ref
    patches: Patches

DocumentPatched = Union[ # type: ignore[misc] # https://github.com/python/mypy/issues/731
    MessageSent,
    ModelChanged,
    ColumnDataChanged,
    ColumnsStreamed,
    ColumnsPatched,
    TitleChanged,
    RootAdded,
    RootRemoved,
]

DocumentChanged = DocumentPatched  # type: ignore[misc] # https://github.com/python/mypy/issues/731

class RootsJson(TypedDict):
    root_ids: List[ID]
    references: List[ReferenceJson]

class DocJson(TypedDict):
    version: str | None
    title: str | None
    defs: List[ModelDef] | None
    roots: RootsJson

class PatchJson(TypedDict):
    events: List[DocumentChanged]
    references: List[ReferenceJson]

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2021, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Organize CSS named colors into useful groups according to general hue.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Bokeh imports
from .util import ColorGroup

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'black',
    'blue',
    'brown',
    'cyan',
    'green',
    'orange',
    'pink',
    'purple',
    'red',
    'white',
    'yellow',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

class black(ColorGroup):
    ''' CSS "Black" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: gainsboro
    .. bokeh-color:: lightgray
    .. bokeh-color:: silver
    .. bokeh-color:: darkgray
    .. bokeh-color:: gray
    .. bokeh-color:: dimgray
    .. bokeh-color:: lightslategray
    .. bokeh-color:: slategray
    .. bokeh-color:: darkslategray
    .. bokeh-color:: black
    '''
    _colors = ('Gainsboro', 'LightGray', 'Silver', 'DarkGray', 'Gray', 'DimGray', 'LightSlateGray',
               'SlateGray', 'DarkSlateGray', 'Black')

class blue(ColorGroup):
    ''' CSS "Blue" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: lightsteelblue
    .. bokeh-color:: powderblue
    .. bokeh-color:: lightblue
    .. bokeh-color:: skyblue
    .. bokeh-color:: lightskyblue
    .. bokeh-color:: deepskyblue
    .. bokeh-color:: dodgerblue
    .. bokeh-color:: cornflowerblue
    .. bokeh-color:: steelblue
    .. bokeh-color:: royalblue
    .. bokeh-color:: blue
    .. bokeh-color:: mediumblue
    .. bokeh-color:: darkblue
    .. bokeh-color:: navy
    .. bokeh-color:: midnightblue
    '''
    _colors = ('LightSteelBlue', 'PowderBlue', 'LightBlue', 'SkyBlue', 'LightSkyBlue', 'DeepSkyBlue', 'DodgerBlue',
               'CornflowerBlue', 'SteelBlue', 'RoyalBlue', 'Blue', 'MediumBlue', 'DarkBlue', 'Navy', 'MidnightBlue')

class brown(ColorGroup):
    ''' CSS "Brown" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: cornsilk
    .. bokeh-color:: blanchedalmond
    .. bokeh-color:: bisque
    .. bokeh-color:: navajowhite
    .. bokeh-color:: wheat
    .. bokeh-color:: burlywood
    .. bokeh-color:: tan
    .. bokeh-color:: rosybrown
    .. bokeh-color:: sandybrown
    .. bokeh-color:: goldenrod
    .. bokeh-color:: darkgoldenrod
    .. bokeh-color:: peru
    .. bokeh-color:: chocolate
    .. bokeh-color:: saddlebrown
    .. bokeh-color:: sienna
    .. bokeh-color:: brown
    .. bokeh-color:: maroon
    '''
    _colors = ('Cornsilk', 'BlanchedAlmond', 'Bisque', 'NavajoWhite', 'Wheat', 'BurlyWood', 'Tan',
               'RosyBrown', 'SandyBrown', 'Goldenrod', 'DarkGoldenrod', 'Peru', 'Chocolate',
               'SaddleBrown', 'Sienna', 'Brown', 'Maroon')

class cyan(ColorGroup):
    ''' CSS "Cyan" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: mediumaquamarine
    .. bokeh-color:: aqua
    .. bokeh-color:: cyan
    .. bokeh-color:: lightcyan
    .. bokeh-color:: paleturquoise
    .. bokeh-color:: aquamarine
    .. bokeh-color:: turquoise
    .. bokeh-color:: mediumturquoise
    .. bokeh-color:: darkturquoise
    .. bokeh-color:: lightseagreen
    .. bokeh-color:: cadetblue
    .. bokeh-color:: darkcyan
    .. bokeh-color:: teal
    '''
    _colors = ('MediumAquamarine', 'Aqua', 'Cyan', 'LightCyan', 'PaleTurquoise', 'Aquamarine', 'Turquoise',
               'MediumTurquoise', 'DarkTurquoise', 'LightSeaGreen', 'CadetBlue', 'DarkCyan', 'Teal')

class green(ColorGroup):
    ''' CSS "Green" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: darkolivegreen
    .. bokeh-color:: olive
    .. bokeh-color:: olivedrab
    .. bokeh-color:: yellowgreen
    .. bokeh-color:: limegreen
    .. bokeh-color:: lime
    .. bokeh-color:: lawngreen
    .. bokeh-color:: chartreuse
    .. bokeh-color:: greenyellow
    .. bokeh-color:: springgreen
    .. bokeh-color:: mediumspringgreen
    .. bokeh-color:: lightgreen
    .. bokeh-color:: palegreen
    .. bokeh-color:: darkseagreen
    .. bokeh-color:: mediumseagreen
    .. bokeh-color:: seagreen
    .. bokeh-color:: forestgreen
    .. bokeh-color:: green
    .. bokeh-color:: darkgreen
    '''
    _colors = ('DarkOliveGreen', 'Olive', 'OliveDrab', 'YellowGreen', 'LimeGreen', 'Lime', 'LawnGreen',
               'Chartreuse', 'GreenYellow', 'SpringGreen', 'MediumSpringGreen', 'LightGreen', 'PaleGreen',
               'DarkSeaGreen', 'MediumSeaGreen', 'SeaGreen', 'ForestGreen', 'Green', 'DarkGreen')

class orange(ColorGroup):
    ''' CSS "Orange" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: orangered
    .. bokeh-color:: tomato
    .. bokeh-color:: coral
    .. bokeh-color:: darkorange
    .. bokeh-color:: orange
    '''
    _colors = ('OrangeRed', 'Tomato', 'Coral', 'DarkOrange', 'Orange')

class pink(ColorGroup):
    ''' CSS "Pink" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: pink
    .. bokeh-color:: lightpink
    .. bokeh-color:: hotpink
    .. bokeh-color:: deeppink
    .. bokeh-color:: palevioletred
    .. bokeh-color:: mediumvioletred
    '''
    _colors = ('Pink', 'LightPink', 'HotPink', 'DeepPink', 'PaleVioletRed', 'MediumVioletRed')

class purple(ColorGroup):
    ''' CSS "Purple" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: lavender
    .. bokeh-color:: thistle
    .. bokeh-color:: plum
    .. bokeh-color:: violet
    .. bokeh-color:: orchid
    .. bokeh-color:: fuchsia
    .. bokeh-color:: magenta
    .. bokeh-color:: mediumorchid
    .. bokeh-color:: mediumpurple
    .. bokeh-color:: blueviolet
    .. bokeh-color:: darkviolet
    .. bokeh-color:: darkorchid
    .. bokeh-color:: darkmagenta
    .. bokeh-color:: purple
    .. bokeh-color:: indigo
    .. bokeh-color:: darkslateblue
    .. bokeh-color:: slateblue
    .. bokeh-color:: mediumslateblue
    '''
    _colors = ('Lavender', 'Thistle', 'Plum', 'Violet', 'Orchid', 'Fuchsia', 'Magenta', 'MediumOrchid',
               'MediumPurple', 'BlueViolet', 'DarkViolet', 'DarkOrchid', 'DarkMagenta', 'Purple', 'Indigo',
               'DarkSlateBlue', 'SlateBlue', 'MediumSlateBlue')

class red(ColorGroup):
    ''' CSS "Red" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: lightsalmon
    .. bokeh-color:: salmon
    .. bokeh-color:: darksalmon
    .. bokeh-color:: lightcoral
    .. bokeh-color:: indianred
    .. bokeh-color:: crimson
    .. bokeh-color:: firebrick
    .. bokeh-color:: darkred
    .. bokeh-color:: red
    '''
    _colors = ('LightSalmon', 'Salmon', 'DarkSalmon', 'LightCoral', 'IndianRed', 'Crimson', 'FireBrick', 'DarkRed', 'Red')

class white(ColorGroup):
    ''' CSS "White" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: white
    .. bokeh-color:: snow
    .. bokeh-color:: honeydew
    .. bokeh-color:: mintcream
    .. bokeh-color:: azure
    .. bokeh-color:: aliceblue
    .. bokeh-color:: ghostwhite
    .. bokeh-color:: whitesmoke
    .. bokeh-color:: seashell
    .. bokeh-color:: beige
    .. bokeh-color:: oldlace
    .. bokeh-color:: floralwhite
    .. bokeh-color:: ivory
    .. bokeh-color:: antiquewhite
    .. bokeh-color:: linen
    .. bokeh-color:: lavenderblush
    .. bokeh-color:: mistyrose
    '''
    _colors = ('White', 'Snow', 'Honeydew', 'MintCream', 'Azure', 'AliceBlue', 'GhostWhite', 'WhiteSmoke', 'Seashell',
               'Beige', 'OldLace', 'FloralWhite', 'Ivory', 'AntiqueWhite', 'Linen', 'LavenderBlush', 'MistyRose')

class yellow(ColorGroup):
    ''' CSS "Yellow" Color Group as defined by https://www.w3schools.com/colors/colors_groups.asp

    .. bokeh-color:: yellow
    .. bokeh-color:: lightyellow
    .. bokeh-color:: lemonchiffon
    .. bokeh-color:: lightgoldenrodyellow
    .. bokeh-color:: papayawhip
    .. bokeh-color:: moccasin
    .. bokeh-color:: peachpuff
    .. bokeh-color:: palegoldenrod
    .. bokeh-color:: khaki
    .. bokeh-color:: darkkhaki
    .. bokeh-color:: gold
    '''
    _colors = ('Yellow', 'LightYellow', 'LemonChiffon', 'LightGoldenrodYellow', 'PapayaWhip',
               'Moccasin', 'PeachPuff', 'PaleGoldenrod', 'Khaki', 'DarkKhaki', 'Gold')

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2021, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' These define the standard error codes and messages for Bokeh
validation checks.

1001 *(BAD_COLUMN_NAME)*
    A glyph has a property set to a field name that does not correspond to any
    column in the |GlyphRenderer|'s data source.

1002 *(MISSING_GLYPH)*
    A |GlyphRenderer| has no glyph configured.

1003 *(NO_SOURCE_FOR_GLYPH)*
    A |GlyphRenderer| has no data source configured.

1004 *(REQUIRED_RANGE)*
    A |Plot| is missing one or more required default ranges (will result in
    blank plot).

1005 *(MISSING_GOOGLE_API_KEY)*
    Google Maps API now requires an API key for all use. See
    https://developers.google.com/maps/documentation/javascript/get-api-key
    for more information on how to obtain your own, to use for the
    ``api_key`` property of your Google Map plot .

1006 *(NON_MATCHING_DATA_SOURCES_ON_LEGEND_ITEM_RENDERERS)*
    All data_sources on ``LegendItem.renderers`` must match when LegendItem.label
    is type field.

1007 *(MISSING_MERCATOR_DIMENSION)*
    ``MercatorTicker`` and ``MercatorTickFormatter``models must have their
    ``dimension`` property set to ``'lat'`` or ``'lon'``.

1008 *(REQUIRED_SCALE)*
    A |Scale| on is missing one or more required default scales (will result in
    blank plot).

1009 *(INCOMPATIBLE_SCALE_AND_RANGE)*
    A |Scale| type is incompatible with one or more ranges on the same plot
    dimension (will result in blank plot).

1010 *(CDSVIEW_SOURCE_DOESNT_MATCH)*
    A |GlyphRenderer| has a ``CDSView`` whose source doesn't match the ``GlyphRenderer``'s
    data source.

1011 *(MALFORMED_GRAPH_SOURCE)*
    The ``GraphSource`` is incorrectly configured.

1012 *(INCOMPATIBLE_MAP_RANGE_TYPE)*
    Map plots can only support ``Range1d`` types, not data ranges.

1013 *(INCOMPATIBLE_POINT_DRAW_RENDERER)*
    The ``PointDrawTool`` renderers may only reference ``XYGlyph`` models.

1014 *(INCOMPATIBLE_BOX_EDIT_RENDERER)*
    The ``BoxEditTool`` renderers may only reference ``Rect`` glyph models.

1015 *(INCOMPATIBLE_POLY_DRAW_RENDERER)*
    The ``PolyDrawTool`` renderers may only reference ``MultiLine`` and ``Patches`` glyph models.

1016 *(INCOMPATIBLE_POLY_EDIT_RENDERER)*
    The ``PolyEditTool`` renderers may only reference ``MultiLine`` and ``Patches`` glyph models.

1017 *(INCOMPATIBLE_POLY_EDIT_VERTEX_RENDERER)*
    The ``PolyEditTool`` vertex_renderer may only reference ``XYGlyph`` models.

1018 *(NO_RANGE_TOOL_RANGES)*
    The ``RangeTool`` must have at least one of ``x_range`` or ``y_range`` configured

1019 *(DUPLICATE_FACTORS)*
    ``FactorRange`` must specify a unique list of categorical factors for an axis.

1020 *(BAD_EXTRA_RANGE_NAME)*
    An extra range name is configured with a name that does not correspond to any range.

1021 *(EQUAL_SLIDER_START_END)*
    ``noUiSlider`` most have a nonequal start and end.

1022 *(MIN_PREFERRED_MAX_WIDTH)*
    Expected min_width <= width <= max_width

1023 *(MIN_PREFERRED_MAX_HEIGHT)*
    Expected min_height <= height <= max_height

1024 *(CDSVIEW_FILTERS_WITH_CONNECTED)*
    ``CDSView`` filters are not compatible with glyphs with connected topology such as Line or Patch.

1025 *(INCOMPATIBLE_LINE_EDIT_RENDERER)*
    The ``LineEditTool`` renderers may only reference ``MultiLine`` and ``Line`` glyph models.

1026 *(INCOMPATIBLE_LINE_EDIT_INTERSECTION_RENDERER)*
    The ``LineEditTool`` intersection_enderer may only reference ``LineGlyph`` models.

1027 *(REPEATED_LAYOUT_CHILD)*
    The same model can't be used multiple times in a layout.

9999 *(EXT)*
    Indicates that a custom error check has failed.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Bokeh imports
from .issue import Error

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

BAD_COLUMN_NAME = Error(
    1001,
    "BAD_COLUMN_NAME",
    "Glyph refers to nonexistent column name. This could either be due to a misspelling or typo, or due to an expected column being missing. ") # NOQA
MISSING_GLYPH = Error(
    1002,
    "MISSING_GLYPH",
    "Glyph renderer has no glyph set")
NO_SOURCE_FOR_GLYPH = Error(
    1003,
    "NO_SOURCE_FOR_GLYPH",
    "Glyph renderer has no data source")
REQUIRED_RANGE = Error(
    1004,
    "REQUIRED_RANGE",
    "A required Range object is missing")
MISSING_GOOGLE_API_KEY = Error(
    1005,
    "MISSING_GOOGLE_API_KEY",
    "Google now requires API keys for all Google Maps usage")
NON_MATCHING_DATA_SOURCES_ON_LEGEND_ITEM_RENDERERS = Error(
    1006,
    "NON_MATCHING_DATA_SOURCES_ON_LEGEND_ITEM_RENDERERS",
    "LegendItem.label is a field, but renderer data sources don't match")
MISSING_MERCATOR_DIMENSION = Error(
    1007,
    "MISSING_MERCATOR_DIMENSION",
    "Mercator Tickers and Formatters must have their dimension property set to 'lat' or 'lon'")
REQUIRED_SCALE = Error(
    1008,
    "REQUIRED_SCALE",
    "A required Scale object is missing")
INCOMPATIBLE_SCALE_AND_RANGE = Error(
    1009,
    "INCOMPATIBLE_SCALE_AND_RANGE",
    "A Scale is incompatible with one or more ranges on the same plot dimension")
CDSVIEW_SOURCE_DOESNT_MATCH = Error(
    1010,
    "CDSVIEW_SOURCE_DOESNT_MATCH",
    "CDSView used by Glyph renderer must have a source that matches the Glyph renderer's data source")
MALFORMED_GRAPH_SOURCE = Error(
    1011,
    "MALFORMED_GRAPH_SOURCE",
    "The GraphSource is incorrectly configured")
INCOMPATIBLE_MAP_RANGE_TYPE = Error(
    1012,
    "INCOMPATIBLE_MAP_RANGE_TYPE",
    "Map plots can only support Range1d types, not data ranges")
INCOMPATIBLE_POINT_DRAW_RENDERER = Error(
    1013,
    "INCOMPATIBLE_POINT_DRAW_RENDERER",
    "PointDrawTool renderers may only reference XYGlyph models")
INCOMPATIBLE_BOX_EDIT_RENDERER = Error(
    1014,
    "INCOMPATIBLE_BOX_EDIT_RENDERER",
    "BoxEditTool renderers may only reference Rect glyph models")
INCOMPATIBLE_POLY_DRAW_RENDERER = Error(
    1015,
    "INCOMPATIBLE_POLY_DRAW_RENDERER",
    "PolyDrawTool renderers may only reference MultiLine and Patches glyph models")
INCOMPATIBLE_POLY_EDIT_RENDERER = Error(
    1016,
    "INCOMPATIBLE_POLY_EDIT_RENDERER",
    "PolyEditTool renderers may only reference MultiLine and Patches glyph models")
INCOMPATIBLE_POLY_EDIT_VERTEX_RENDERER = Error(
    1017,
    "INCOMPATIBLE_POLY_EDIT_VERTEX_RENDERER",
    "PolyEditTool vertex_renderer may only reference XYGlyph models")
NO_RANGE_TOOL_RANGES = Error(
    1018,
    "NO_RANGE_TOOL_RANGES",
    "RangeTool must have at least one of x_range or y_range configured")
DUPLICATE_FACTORS = Error(
    1019,
    "DUPLICATE_FACTORS",
    "FactorRange must specify a unique list of categorical factors for an axis")
BAD_EXTRA_RANGE_NAME = Error(
    1020,
    "BAD_EXTRA_RANGE_NAME",
    "An extra range name is configured with a name that does not correspond to any range")
EQUAL_SLIDER_START_END = Error(
    1021,
    "EQUAL_SLIDER_START_END",
    "Slider 'start' and 'end' cannot be equal")
MIN_PREFERRED_MAX_WIDTH = Error(
    1022,
    "MIN_PREFERRED_MAX_WIDTH",
    "Expected min_width <= width <= max_width")
MIN_PREFERRED_MAX_HEIGHT = Error(
    1023,
    "MIN_PREFERRED_MAX_HEIGHT",
    "Expected min_height <= height <= max_height")
CDSVIEW_FILTERS_WITH_CONNECTED = Error(
    1024,
    "CDSVIEW_FILTERS_WITH_CONNECTED",
    "CDSView filters are not compatible with glyphs with connected topology such as Line or Patch")
INCOMPATIBLE_LINE_EDIT_RENDERER = Error(
    1025,
    "INCOMPATIBLE_LINE_EDIT_RENDERER",
    "LineEditTool renderers may only reference MultiLine and Line glyph models")
INCOMPATIBLE_LINE_EDIT_INTERSECTION_RENDERER = Error(
    1026,
    "INCOMPATIBLE_LINE_EDIT_INTERSECTION_RENDERER",
    "LineEditTool intersection_renderer may only reference LineGlyph models")
REPEATED_LAYOUT_CHILD = Error(
    1027,
    "REPEATED_LAYOUT_CHILD",
    "The same model can't be used multiple times in a layout")
EXT = Error(
    9999,
    "EXT",
    "Custom extension reports error")

__all__ = ()

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2021, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Encapulate the management of Document models with a DocumentModelManager
class.

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
import contextlib
import weakref
from typing import (
    TYPE_CHECKING,
    Dict,
    Generator,
    Iterator,
    List,
    Set,
)

# Bokeh imports
from ..core.types import ID
from ..model import Model
from ..util.datatypes import MultiValuedDict

if TYPE_CHECKING:
    from .document import Document

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'DocumentModelManager',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

class DocumentModelManager:
    ''' Manage and provide access to all of the models that belong to a Bokeh
    Document.

    The set of "all models" means specifically all the models reachable from
    references form a Document's roots.

    '''

    _document : weakref.ReferenceType[Document]
    _freeze_count: int
    _models: Dict[ID, Model]
    _models_by_name: MultiValuedDict[str, Model]
    _seen_model_ids: Set[ID] = set()

    def __init__(self, document: Document):
        '''

        Args:
            document (Document): A Document to manage models for
                A weak reference to the Document will be retained

        '''
        self._document = weakref.ref(document)
        self._freeze_count = 0
        self._models = {}
        self._models_by_name = MultiValuedDict()
        self._seen_model_ids = set()

    def __len__(self) -> int:
        return len(self._models)

    def __getitem__(self, id: ID) -> Model:
        return self._models[id]

    def __setitem__(self, id: ID, model: Model) -> None:
        self._models[id] = model

    def __contains__(self, id: ID) -> bool:
        return id in self._models

    def __iter__(self) -> Iterator[Model]:
        return iter(self._models.values())

    def destroy(self) -> None:
        ''' Clean up references to the Documents models

        '''

        # probably better to implement a destroy protocol on models to
        # untangle everything, then the collect below might not be needed
        for m in self._models.values():
            m.destroy()

        del self._models
        del self._models_by_name

    @contextlib.contextmanager
    def freeze(self) -> Generator[None, None, None]:
        ''' Defer expensive model recompuation until intermediate updates are
        complete.

        Making updates to the model graph might trigger events that cause more
        updates. This context manager can be used to prevent expensive model
        recompuation from happening until all events have finished and the
        Document state is quiescent.

        Example:

        .. code-block:: python

            with models.freeze():
                # updates that might change the model graph, that might trigger
                # updates that change the model graph, etc. Recompuation will
                # happen once at the end.

        '''
        self._push_freeze()
        yield
        self._pop_freeze()

    def get_all_by_name(self, name: str) -> List[Model]:
        ''' Find all the models for this Document with a given name.

        Args:
            name (str) : the name of a model to search for

        Returns
            A list of models

        '''
        return self._models_by_name.get_all(name)

    def get_by_id(self, id: ID) -> Model | None:
        ''' Find the model for this Document with a given ID.

        Args:
            id (ID) : model ID to search for
                If no model with the given ID exists, returns None

        Return:
            a Model or None

        '''
        return self._models.get(id, None)

    def get_one_by_name(self, name: str) -> Model | None:
        ''' Find a single model for this Document with a given name.

        If multiple models are found with the name, an error is raised.

        Args:
            name (str) : the name of a model to search for

        Returns
            A model with the given name, or None

        '''
        return self._models_by_name.get_one(name, f"Found more than one model named '{name}'")

    def invalidate(self) -> None:
        ''' Recompute the set of all models, if not currently frozen

        Returns:
            None

        '''
        if self._freeze_count == 0:  # only recompute when competely unfrozen
            self.recompute()

    def recompute(self) -> None:
        ''' Recompute the set of all models based on references reachable from
        the Document's current roots.

        This computation can be expensive. Use ``freeze`` to wrap operations
        that update the model object graph to avoid over-recompuation

        .. note::
            Any models that remove during recomputation will be noted as
            "previously seen"

        '''
        document = self._document()
        if document is None:
            return

        new_models: Set[Model] = set()
        for mr in document.roots:
            new_models |= mr.references()

        old_models = set(self._models.values())

        to_detach = old_models - new_models
        to_attach = new_models - old_models

        recomputed: Dict[ID, Model] = {}
        recomputed_by_name: MultiValuedDict[str, Model] = MultiValuedDict()

        for mn in new_models:
            recomputed[mn.id] = mn
            if mn.name is not None:
                recomputed_by_name.add_value(mn.name, mn)

        for md in to_detach:
            self._seen_model_ids.add(md.id)
            md._detach_document()

        for ma in to_attach:
            ma._attach_document(document)

        self._models = recomputed
        self._models_by_name = recomputed_by_name

    # XXX (bev) In theory, this is a potential issue for long-running apps that
    # update the model graph continuously, since this set of "seen" model ids can
    # grow without bound.
    def seen(self, id: ID) -> bool:
        ''' Report whether a model id has ever previously belonged to this
        Document.

        Args:
            id (ID) : the model id of a model to check

        Returns:
            bool

        '''
        return id in self._seen_model_ids

    def update_name(self, model: Model, old_name: str | None, new_name: str | None) -> None:
        ''' Update the name for a model.

        .. note::
            This function and the internal name mapping exist to support
            optimizing the common case of name lookup for models. Keeping a
            dedicated name index is faster than using generic ``bokeh.query``
            functions with a name selector

        Args:
            model (Model) : a model to update the name for

            old_name(str, None) : a previous name for the model, or None

            new_name(str, None) : a new name for the model, or None

        Returns:
            None

        '''
        if old_name is not None:
            self._models_by_name.remove_value(old_name, model)
        if new_name is not None:
            self._models_by_name.add_value(new_name, model)

    def _push_freeze(self) -> None:
        self._freeze_count += 1

    def _pop_freeze(self) -> None:
        self._freeze_count -= 1
        if self._freeze_count == 0:
            self.recompute()

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2021, Anaconda, Inc., and Bokeh Contributors.
# All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
#-----------------------------------------------------------------------------
''' Define a Pytest plugin for a log file fixture

'''

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import annotations

import logging # isort:skip
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterator,
    List,
    NoReturn,
    Sequence,
)
from warnings import warn

# External imports
import pytest

if TYPE_CHECKING:
    import py
    from _pytest import config, nodes
    from selenium.webdriver.remote.webdriver import WebDriver

#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

__all__ = (
    'driver',
    'has_no_console_errors',
    'pytest_report_collectionfinish',
)

#-----------------------------------------------------------------------------
# General API
#-----------------------------------------------------------------------------

def pytest_report_collectionfinish(config: config.Config, startdir: py.path.local, items: Sequence[nodes.Item]) -> List[str]:
    '''

    '''
    driver_name: str = config.getoption('driver', 'chrome').lower()
    asserts = "ON" if driver_name == "chrome" else "OFF"
    return ["", f"Bokeh selenium tests using {driver_name!r} driver (no-console-error assertions: {asserts})"]

@pytest.fixture(scope="session")
def driver(pytestconfig: config.Config) -> Iterator[WebDriver]:
    ''' Select and configure a Selenium webdriver for integration tests.

    '''
    driver_name: str = pytestconfig.getoption('driver', 'chrome').lower()

    def chrome() -> WebDriver:
        from selenium.webdriver.chrome.options import Options
        from selenium.webdriver.chrome.webdriver import WebDriver as Chrome
        options = Options()
        options.add_argument("--headless")
        options.add_argument("--no-sandbox")
        options.add_argument("--window-size=1920x1080")
        return Chrome(options=options)

    def firefox() -> WebDriver:
        from selenium.webdriver.firefox.options import Options
        from selenium.webdriver.firefox.webdriver import WebDriver as Firefox
        options = Options()
        options.add_argument("--headless")
        options.add_argument("--window-size=1920x1080")
        return Firefox(options=options)

    def safari() -> WebDriver:
        from selenium.webdriver.safari.webdriver import WebDriver as Safari
        return Safari()

    driver: WebDriver
    if driver_name == "chrome":
        driver = chrome()
    elif driver_name == "firefox":
        driver = firefox()
    elif driver_name == "safari":
        driver = safari()
    else:
        raise ValueError("expected 'chrome', 'firefox' or 'safari'")

    driver.implicitly_wait(10)
    yield driver
    driver.quit()

@pytest.fixture(scope="session")
def has_no_console_errors(pytestconfig: config.Config) -> Callable[[WebDriver], bool | NoReturn]:
    ''' Provide a function to assert no browser console errors are present.

    Unfortunately logs are only accessibly with Chrome web driver, see e.g.

        https://github.com/mozilla/geckodriver/issues/284

    For non-Chrome webdrivers this check always returns True.

    '''
    driver_name: str = pytestconfig.getoption('driver').lower()

    if driver_name == "chrome":

        def func(driver: WebDriver) -> bool | NoReturn:
            logs = driver.get_log('browser')
            severe_errors = [x for x in logs if x.get('level') == 'SEVERE']
            non_network_errors = [l for l in severe_errors if l.get('type') != 'network']

            if len(non_network_errors) == 0:
                if len(severe_errors) != 0:
                    warn(f"There were severe network errors (this may or may not have affected your test): {severe_errors}")
                return True

            # XXX: no return should be needed with NoReturn type (type-checker bug?)
            return pytest.fail(f"Console errors: {non_network_errors}")

    else:
        def func(driver: WebDriver) -> bool | NoReturn:
            return True

    return func

#-----------------------------------------------------------------------------
# Dev API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Private API
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Code
#-----------------------------------------------------------------------------
